#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ltv_mpc']

package_data = \
{'': ['*']}

install_requires = \
['qpsolvers >=1.8.0']

extras_require = \
{'doc': ['sphinx']}

setup(name='ltv-mpc',
      version='0.7.0',
      description='Linear time-variant model predictive control in Python.',
      author=None,
      author_email='Stéphane Caron <stephane.caron@normalesup.org>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )
